/***************************************************************************
 *   Copyright (C) 2006 by Julien Keller                                   *
 *   j.keller@imperial.ac.uk                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>

#include "math.h"
#include <complex>
#include "time.h"


#define Complex  std::complex<double>
#define _Complex_I std::complex<double>(0, 1)
#define _1  (std::complex<double>(1, 0))
#define _2  (std::complex<double>(2, 0))
#define _3  (std::complex<double>(3, 0))
#define _0  (std::complex<double>(0, 0))
#define PI  3.14159265358979323846

#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)
#define pow5(x) (x*pow4(x))
#define pow6(x) pow2(pow3(x))
#define pow8(x) pow2(pow4(x))
   


/* Packages to handle matrices operations */
#include "m_matrix.h"

// Default variables //
int choice=0;              // 0 -> normaL method 1 => refined method 2=>experimental version ?
int K=6;                   //H^0(L^K)
int method_compute_vol=1;  // 1-> normal = improved , 2-> not improved along branch 

//cut off parameters fixed arbitrarily--
double parameter_smallchart=pow(0.72,6); //.117
double parameter_bigchart=pow(0.905,6);  //.531
double parameter1_FNcut=0.95;
double parameter2_FNcut=1.05;
double parameter3_FNcut=0.05;
double branch=4.5;double brep=0.01; 
//--------------------------------------

// Matrices containing the FS metric //
int size_B;
int size_A;
MMatrix<double> A;
MMatrix<double>A2;
MMatrix<double> B;
MMatrix<double>B2;
MMatrix<double> C_A;   //non diagonal terms for A
MMatrix<double> C_A2;
MMatrix<double> new_C_A2;
MMatrix<double> C_B;  //non diagonal terms for B
MMatrix<double> C_B2;
MMatrix<double> update_new_C_A;
MMatrix<Complex> update_new_C_A2;
MMatrix<double> update_new_C_B;
//---------------------------------//


// Variables to do the computation of volume form //
MMatrix<Complex> Z;
MMatrix<Complex> Deriv_x;
MMatrix<Complex> Deriv_y;
MMatrix<Complex> Zw;
MMatrix<Complex> Deriv_xw;
MMatrix<Complex> Deriv_yw;


int MAX_RANDOM=200;     //used for building random matrices
std::string name_file;  //name of file for output

int write_results;      // =1 write the results in a file
int random_matrices;    //=1 random matrices, =0 read matrices in file



MMatrix <int> Nondiagonalterms_A; //couples of (i,j) for x^i y^j non trivial that give non diagonal terms
MMatrix <int> Nondiagonalterms_B; // C_A and C_B
MMatrix <int> Nondiagonalterms_A2;
MMatrix <int> Diagonalterms_A; // (i,j) of diagonal terms where (i,j)=(diagonalterms[0],..[1])
                               // form the basic data for A matrix.
MMatrix <int> Diagonalterms_A2;
MMatrix <int> Diagonalterms_B;
MMatrix <int> Inverseterms_A;  // each line contains the i of C_i forming a 3x3 matrices to inverse
MMatrix <int> Inverseterms_A2;
MMatrix <int> Inverseterms_B2;
MMatrix <int> Inverseterms_C_A; // each line contains the i of C_i forming the 3x3 matrice i to inverse 
MMatrix <int> Inverseterms_C_B;
MMatrix <int> Matrices_Inverseterms_A;
MMatrix <int> Matrices_Inverseterms_B;

/* radius of charts */
double rad=pow(parameter2_FNcut,6.) * 2.;
double radius2=1.75; //best value :1.65;
//size of charts by default //
double Numx=20.;
double Nump=20.;
double Numv=14.;
double Numw=14.;


//-------------------------//
double maxval,minval;
int number=0;         /* number of iterations by default*/
int itk=0;
time_t seconds;      /* time of computation */
double VOL=5.40389; // \int d\mu / \int d\nu = 2*72\pi^2 / 263.
double ernorm,fubmass;
double distspace=3, numbdown=3;  //distribution spacing,numberdown
double lowbound = -numbdown * distspace;
double speed=1;  // constant Kappa. Must be not too large

double Euler_density=0;
double Euler_density_Temp=0;

/* Characteristic function for the charts covering  -- cut off functions */
double FNBigcut(double,double);
double FNCut (double,double,double);

/* Compute D=\sum G^{\alpha,\beta}z_{\alpha}\bar{z}_{\beta} function locally */
void summand(double,double,double,Complex,Complex,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,double,double);

/* Fix size of matrices, read data from default file "param"  or other file specified on command line,
 and set the matrices A,B,C_A,C_B,C_A_B */
void initialisation(int,MMatrix<double>&,MMatrix<double>&,int,char**);

/* Compute mass,new_A,new_B,.... for first chart */
void firstchart(double&,MMatrix<double>&,MMatrix<double>&,int,MMatrix<double>&,MMatrix<double>&,double&,MMatrix<double> &);

/* Compute mass,new_A,new_B,.... for second chart */
void secondchart(double&,MMatrix<double>&,MMatrix<double>&,int,MMatrix<double>&,MMatrix<double>&,double&,MMatrix<double>&);


/* Differences between the computation and last one */
void compute_distance(int,MMatrix<double>&,MMatrix<double>&);


/* Inverse data for normal method (balanced metric) */
void data_inversion(MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&);


/* Inverse data for refined method */
void data_inversion2(MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&);


/* Inverse data using direct computation of inverse 3x3 matrices and computation
of best kappa constant */
void data_inversion3(MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&);


/* The following procedures computes the volume using the FS metric, i.e basically
find a determinant. The second procedure does it in the most naive way */
void compute_volume(MMatrix<double> &, MMatrix<double> &, MMatrix<double>&, MMatrix<double>&, Complex,Complex,double,double&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&);

void compute_volume2(MMatrix<double> &, MMatrix<double> &, MMatrix<double>&, MMatrix<double>&, Complex,Complex,double,double&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,double);

/* Computation of hermitian scalar product */
void hermit(MMatrix<double>&,MMatrix<double>&,MMatrix<Complex>&,MMatrix<Complex>&,Complex&,Complex&);
void hermit2(MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<double>&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,MMatrix<Complex>&,Complex&,Complex&);


/* skip end of line when reading file */
void skipline(ifstream &);

/* write the results of computation in a file */
void write_results_in_file(double);

/* create matrices for any K -- Generate the matrices containing the coordinates H_ij z^iz^j
to compute and procedures to handle them */
void create_matrices(int);
void generate_non_diagonal_terms(int);
void generate_diagonal_terms(int);
void generate_inverse_terms(int);
void fill_Nondiagonalterms_A(int,int);
void fill_Nondiagonalterms_B(int,int);
void generate_non_diagonal_terms2(int);	
bool present(int*, MMatrix<int>);
		
/* Symmetry in the variables */
template <typename T>
void symmetry(MMatrix<T> &);


