/***************************************************************************
 *   Copyright (C) 2006 by Julien Keller                                   *
 *   j.keller@imperial.ac.uk                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "m_matrix.h"
#include <complex>
#include "MMatrix.h"  //Contains library to handle quickly matrices in memory


#define Complex  std::complex<double>
#include <iostream>

/* This is the library NEWMAT for inverse, determinant , etc 
containing algebraic operations */

#define WANT_STREAM                  // include.h will get stream fns
#define WANT_MATH                    // include.h will get math fns
                                     // newmatap.h will get include.h
#include "newmatap.h"                // need matrix applications
#include "newmatio.h"                // need matrix output routines
#ifdef use_namespace
using namespace NEWMAT;              // access NEWMAT namespace
#endif
/*****************************************/


/* Only show the matrix */
void show(MMatrix<double> A) {
int i, j;
	printf("showing... \n");
	for(i=0;i<A.getHeight();i++) {
		for(j=0;j<A.getWidth();j++) {
		        if (A[i][j]>=1000 || -A[i][j]>=1000)
			{
			    printf("%+.2e ", A[i][j]);
			}
			else
			{ if     
			  (A[i][j]>=100 || -A[i][j]>=100){
		           printf("%+7.4f ", A[i][j]);
			    }
			  else
			  {  
			  if     
			  (A[i][j]>=10 || -A[i][j]>=10){
		           printf("%+7.5f ", A[i][j]);
			    }
			   else 
			   {  
			     if (A[i][j]==0)
			        { printf("    0     "); } 
				else 
				  {
				  	   printf("%+6.6f ", A[i][j]);
				  }	   
			   }
			  }
			}        
			
		}
		printf("\n");
	}
}
void show(MMatrix<int> A) {
	int i, j;
	printf("showing... \n");
	for(i=0;i<A.getHeight();i++) {
		for(j=0;j<A.getWidth();j++) {
			
			printf(" %d ",A[i][j]);        
			
		}
		printf("\n");
	}
}



/* Show the matrix && a string */
void show(MMatrix <double> A,char* a) {
int i, j;
    if((A.getHeight()!=0) && (A.getWidth()!=0)){
        printf("showing...%s \n",a);
        for(i=0;i<A.getHeight();i++) {
            for(j=0;j<A.getWidth();j++) {
                    if (A[i][j]>=1000 || -A[i][j]>=1000)
                {
                    printf("%+.2e ", A[i][j]);
                }
                else
                { if     
                (A[i][j]>=100 || -A[i][j]>=100){
                    printf("%+7.4f ", A[i][j]);
                    }
                else
                {  
                if     
                (A[i][j]>=10 || -A[i][j]>=10){
                    printf("%+7.5f ", A[i][j]);
                    }
        /*		   if (A[i][j]==0)
                    { printf("    0    "); } */
                else 
                {  
                    if (A[i][j]==0)
                        { printf("    0     "); } 
                    else 
                    {
                        printf("%+6.6f ", A[i][j]);
                    }	   
                }
                }
                }        
                
            }
            printf("\n");
        }
    }
}



/* Extending previous procedure to Complex data */
void show(MMatrix<Complex> A) {
int i, j;
	printf("showing... \n");
	for(i=0;i<A.getHeight();i++) {
		for(j=0;j<A.getWidth();j++) {
		if (A[i][j]!=complex<double>(0,0))
		{
		    printf("%.4f %+.4fI ",real(A[i][j]),imag(A[i][j]));
		}     
		else 
		{
			printf("       0       ");
		}	
		}
		printf("\n");
		
	}
}

/* Inverse the matrix */
MMatrix<double> inverse(MMatrix<double> M) {
int i,j;
Matrix Temp(M.getHeight(),M.getWidth());
MMatrix<double> M2(M.getHeight(),M.getWidth());

       for(i=0;i<M.getHeight();i++) {
               for(j=0;j<M.getWidth();j++) {
	                Temp.element(i,j)=M[i][j];
	       }
       }
       Temp=Temp.i();
       for(i=0;i<M.getHeight();i++) {
               for(j=0;j<M.getWidth();j++) {
	                M2[i][j]=Temp.element(i,j);
	       }
       }
       return M2;

}

/* Put 0 everywhere */
void reinitialise(MMatrix<double> &M) {
	M.fillWith(0);
}


/* inverse elements for half of the matrix */
MMatrix<double> inverse_elements(MMatrix<double> M){
int i,j;
  for(i=0;i<M.getHeight();i++)
  {
      for(j=0;j<M.getWidth()-i;j++)
      {
          M[i][j]=1./M[i][j];
      }
  }
  return M;
}


/* normal transpose */
MMatrix<double> transpose(MMatrix<double>M)
{
  return M.transpose();
} 

/* special ""transpose"" */
MMatrix<double> transpose2(MMatrix<double> M){
int i,j;
  MMatrix<double> N(M.getHeight(),M.getWidth());
  for(i=0;i<M.getHeight();i++)
  {
     for(j=0;j<M.getWidth()-i;j++)
     {
        N[i][j]=M[M.getHeight()-1-(i+j)][i]+M[j][M.getWidth()-1-(i+j)];
     }
  }
  return N;
}

/* scalar product of two vectors */	
double  scalar_product(MMatrix<double> M1,MMatrix<double> M2){
int i;
double res=0;
    for (i=0;i<M1.getHeight();i++)
    {
        res+=M1[i][0]*M2[i][0];
    }
    return res;
}

/* Simple procedure to multiply 2 matrices */
MMatrix<double> multiply(MMatrix<double> A,MMatrix<double> B){
int i,j,k;
double sum;
	if(A.getWidth()==B.getHeight()) {
		MMatrix<double> C(A.getHeight(),B.getWidth());
		for(i=0;i<A.getHeight();i++) {
			for(j=0;j<B.getWidth();j++) {
				sum=0.0;
				for(k=0;k<A.getWidth();k++){
					sum+=(A[i][k]*B[k][j]);
				}
				C[i][j]=sum;
			}
		}
		return C;
	}
}

void mult(MMatrix<double> &A,MMatrix<double> B){
int i,j,k;
double sum;
	if(A.getWidth()==B.getHeight()) {
		
		for(i=0;i<A.getHeight();i++) {
			for(j=0;j<B.getWidth();j++) {
				sum=0.0;
				for(k=0;k<A.getWidth();k++) {
					sum+=(A[i][k]*B[k][j]);
				}
				A[i][j]=sum;
			}
		}
	}
}

/* Determinant using NEWMAT */
double determinant(MMatrix<double> M){
int i,j;
Matrix Temp(M.getHeight(),M.getWidth());

       for(i=0;i<M.getHeight();i++) {
               for(j=0;j<M.getWidth();j++) {
	                Temp.element(i,j)=M[i][j];
	       }
       }
       return Temp.Determinant();
}

/* need this for complex 3x3 matrices */
Complex determinant3x3(MMatrix<Complex> M){
int i,j;
return (M[0][0]*M[1][1]*M[2][2]+M[0][1]*M[1][2]*M[2][0]+M[1][0]*M[2][1]*M[0][2]
		-M[2][0]*M[1][1]*M[0][2]-M[0][1]*M[1][0]*M[2][2]-M[0][0]*M[2][1]*M[1][2]);

}

/* this is not the radius of the spectrum but an upper bound */
double spectral_radius(MMatrix<double> M){
	MMatrix<double> U(M.getHeight(),1);
	MMatrix<double> Ut(M.getHeight(),1);
	int i;
	double sum=0;
	U[0][0]=1.;
	U[1][0]=0;
	U[2][0]=0;
	for(i=0;i<40;i++){
		Ut=multiply(M,U);
		U=(1./(Ut.norm2()))*Ut;
	}
	Ut=multiply(M,U);

	for(i=0;i<M.getHeight();i++)
	{
		sum=sum+Ut[i][0]/U[i][0];
	}
	return(abs(sum/M.getHeight()));
	
}



void MJacobi(MMatrix <double> M,MMatrix <double> &D_M){
	SymmetricMatrix S1(M.getHeight());
	DiagonalMatrix D1(M.getHeight());
	int i,j;
	for (i=0;i<M.getHeight();i++){
		
		for(j=0;j<M.getWidth();j++)
	    {
			S1.element(i,j)=M[i][j];
		}
	}
	Jacobi(S1,D1);
	D_M.resize(M.getHeight(),M.getWidth());
	for (i=0;i<M.getHeight();i++){
		D_M[i][i]=D1.element(i,i);
		}
	
}

